@@ -5,11 +5,12 @@ import itertools
 import logging
 import os
 import pathlib
+import queue
 import stat
 import threading
 import typing
 from collections import defaultdict
-from concurrent.futures import ThreadPoolExecutor, Executor, Future
+from concurrent.futures import Future
 
 import click
 from tqdm import tqdm
@@ -52,35 +53,87 @@ class FatalCrawlException(Exception):
     pass
 
 
-class PotentiallySingleThreadedExecutor(ThreadPoolExecutor):
-    def __init__(self, max_workers: int, *args, **kwargs):
-        self.disable = True
-        if max_workers > 1:
-            self.disable = False
-            super().__init__(max_workers, *args, **kwargs)
+class FutureFreeThreadPool:
+    # TODO: benchmark https://github.com/brmmm3/fastthreadpool
+    def __init__(self, threads: int, queue_size: int = 2048):
+        assert threads > 0
+        self.work_queue = queue.Queue(maxsize=queue_size)
+        self.threads = [threading.Thread(target=self._worker_run, daemon=True)
+                        for _ in range(threads)]
+        for t in self.threads:
+            t.start()
 
-    def submit(self, fn, *args, **kwargs):
-        if self.disable:
-            f = Future()
-            f.set_result(fn(*args, **kwargs))
-            return f
-        else:
-            return super().submit(fn, *args, **kwargs)
+    def wait_until_complete(self):
+        self.work_queue.join()
 
-    def map(self, fn, *iterables):
-        if self.disable:
-            return list(map(fn, *iterables))
-        else:
-            return super().map(fn, *iterables)
+    def __enter__(self):
+        return self
 
     def __exit__(self, *args):
-        if not self.disable:
-            return super().__exit__(*args)
+        self.work_queue.join()
+        for _ in self.threads:
+            self.work_queue.put(None)
+        for t in self.threads:
+            t.join()
+
+    @staticmethod
+    def _run_task(fn_args):
+        fn, cb, args = fn_args
+        result = fn(*args)
+        if cb:
+            cb(result)
+
+    def _worker_run(self):
+        try:
+            while True:
+                fn_args = self.work_queue.get(block=True)
+                if fn_args is None:
+                    return
+                self._run_task(fn_args)
+                self.work_queue.task_done()
+                # allow GC
+                del fn_args
+        except Exception:
+            logger.exception('Unhandled error -- exit required')
+            exit(2)
+
+    def submit(self, fn, *args, callback=None):
+        if len(self.work_queue.queue) > 1024:
+            self._run_task((fn, callback, args))
+        else:
+            self.work_queue.put((fn, callback, args))
+
+    def map_unordered(self, fn, iterables):
+        size = len(iterables)
+        result = []
+        complete = threading.Event()
+
+        def map_callback(out):
+            result.append(out)
+            if size == len(result):
+                complete.set()
+
+        for args in iterables:
+            self.submit(fn, *args, callback=map_callback)
+        complete.wait()
+        return result
+
+    def map(self, fn, iterables):
+        wrapped = functools.partial(self._wrap_with_index, fn=fn)
+        iterables = list(enumerate(iterables))
+
+        result = self.map_unordered(wrapped, iterables)
+        result.sort(key=lambda ab: ab[0])
+        return [r for _, r in result]
+
+    @staticmethod
+    def _wrap_with_index(i, args, fn):
+        return i, fn(args)
 
 
 class FileCrawler:
     def __init__(self, roots: typing.List[pathlib.Path],
-                 pool: Executor,
+                 pool: FutureFreeThreadPool,
                  progress: typing.Optional[tqdm] = None):
         assert roots
         if not all(r.is_dir() for r in roots):
@@ -98,13 +151,16 @@ class FileCrawler:
         self.num_directories = len(roots)
 
     def _traverse_path(self, path: str) -> typing.List[Future]:
-        futures = []
         next_dir = None
         for p in os.scandir(path):
             ppath = p.path
             if self._prog is not None:
                 self._prog.update(1)
-            pstat = p.stat(follow_symlinks=False)
+            try:
+                pstat = p.stat(follow_symlinks=False)
+            except OSError as e:
+                logger.error('Unable to stat %s due to: %s', ppath, e)
+                continue
             if pstat.st_dev != self._root_device:
                 logger.debug('Skipping file on separate device %s', ppath)
                 continue
@@ -125,16 +181,16 @@ class FileCrawler:
                 if not next_dir:
                     next_dir = ppath
                 else:
-                    futures.append(self._pool.submit(self._traverse_path, ppath))
+                    self._pool.submit(self._traverse_path, ppath)
             else:
                 logger.debug('Skipping device/socket/unknown: %s', ppath)
         if next_dir:
-            futures.extend(self._traverse_path(next_dir))
-        return futures
+            self._traverse_path(next_dir)
 
     def traverse(self):
-        futures = [self._traverse_path(str(r)) for r in set(self._roots)]
-        self._unwrap_futures(itertools.chain.from_iterable(futures))
+        for r in sorted(set(self._roots)):
+            self._traverse_path(str(r))
+        self._pool.wait_until_complete()
 
     def size_bytes(self):
         return sum(k * len(v) for k, v in self._size_to_paths.items())
@@ -156,8 +212,7 @@ class FileCrawler:
 
 
 class DupeFinder:
-    def __init__(self, size_bytes: int,
-                 pool: Executor,
+    def __init__(self, pool: FutureFreeThreadPool,
                  output: typing.Optional[typing.TextIO] = None,
                  file_progress: typing.Optional[tqdm] = None,
                  byte_progress: typing.Optional[tqdm] = None):
@@ -165,25 +220,24 @@ class DupeFinder:
         self._output = output
         self._file_progress = file_progress
         self._byte_progress = byte_progress
-        self.size_bytes = size_bytes
 
-    def find(self, paths: typing.List[str]) -> typing.List[str]:
+    def find(self, size_bytes: int, paths: typing.List[str]) -> typing.Tuple[str]:
         files_start = len(paths)
-        if self.size_bytes >= 8192:
-            path_groups = list(self._split_by_boundaries(paths))
+        if size_bytes > 8192:
+            path_groups = list(self._split_by_boundaries(size_bytes, paths))
         else:
             path_groups = [paths]
         files_end = sum(len(p) for p in path_groups)
         if self._file_progress is not None:
             self._file_progress.update(files_start - files_end)
         if self._byte_progress is not None:
-            self._byte_progress.update(self.size_bytes * (files_start - files_end))
+            self._byte_progress.update(size_bytes * (files_start - files_end))
 
         dupes = []
         for paths in path_groups:
             hashes = dict()
             paths.sort(key=len)
-            for p, hash in zip(paths, self._map(sha256sum, paths)):
+            for p, hash in zip(paths, self._map(sha256sum, paths, size_bytes)):
                 if not hash:
                     # some read error occurred
                     continue
@@ -202,32 +256,35 @@ class DupeFinder:
                 if self._file_progress is not None:
                     self._file_progress.update(1)
                 if self._byte_progress is not None:
-                    self._byte_progress.update(self.size_bytes)
-        return dupes
+                    self._byte_progress.update(size_bytes)
+        return tuple(dupes)
 
     @staticmethod
     @none_if_io_error
-    def _read_boundary(path: str, size: int) -> bytes:
+    def _read_boundary(path: str, size: int, use_hash=True) -> typing.Union[bytes, int]:
         with open(path, 'rb', buffering=0) as f:
             if size < 0:
                 f.seek(size, os.SEEK_END)
-            return f.read(abs(size))
+            boundary = f.read(abs(size))
+        if use_hash:
+            return hash(boundary)
+        return boundary
 
-    def _split_by_boundary(self, group: typing.List[str], end: bool):
+    def _split_by_boundary(self, size_bytes: int, group: typing.List[str], end: bool):
         matches = defaultdict(list)
         size = -4096 if end else 4096
-        boundaries = self._map(functools.partial(self._read_boundary, size=size), group)
+        boundaries = self._map(functools.partial(self._read_boundary, size=size), group, size_bytes)
         for path, edge in zip(group, boundaries):
             matches[edge].append(path)
-        return [g for boundary, g in matches.items() if boundary and len(g) > 1]
+        return [g for boundary, g in matches.items() if boundary is not None and len(g) > 1]
 
-    def _split_by_boundaries(self, group: typing.List[str]):
-        groups = self._split_by_boundary(group, end=False)
+    def _split_by_boundaries(self, size_bytes: int, group: typing.List[str]):
+        groups = self._split_by_boundary(size_bytes, group, end=False)
         for g in groups:
-            yield from self._split_by_boundary(g, end=True)
+            yield from self._split_by_boundary(size_bytes, g, end=True)
 
-    def _map(self, fn, iterables):
-        if len(iterables) < 16 or self.size_bytes > 2 ** 20 * 32:
+    def _map(self, fn, iterables, size_bytes):
+        if len(iterables) < 16 or size_bytes > 2 ** 20 * 32:
             return map(fn, iterables)
         return self._pool.map(fn, iterables)
 
@@ -235,12 +292,12 @@ class DupeFinder:
 @click.command(help="A duplicate file finder that may be faster in environments with "
                     "millions of files and terabytes of data or over high latency filesystems.")
 @click.argument('input_paths', type=click.Path(
-    exists=True, file_okay=False, writable=True, readable=True), nargs=-1)
+    exists=True, file_okay=False, readable=True), nargs=-1)
 @click.option('--output', type=click.File('w'),
               help='Save null-delimited input/duplicate filename pairs. For stdout use "-".')
 @click.option('--verbose', is_flag=True, help='Enable debug logging.')
 @click.option('--progress', is_flag=True, help='Enable progress bars.')
-@click.option('--concurrency', type=click.IntRange(min=1), default=16,
+@click.option('--concurrency', type=click.IntRange(min=1), default=8,
               help='Level of IO concurrency. Setting to 0 disables all threading/concurrency.')
 def main(input_paths, output, verbose, progress, concurrency):
     input_paths = [pathlib.Path(p) for p in input_paths]
@@ -254,7 +311,7 @@ def main(input_paths, output, verbose, progress, concurrency):
     time_start = datetime.datetime.now()
     with tqdm(smoothing=0, desc='Traversing', unit=' files',
               disable=not progress, mininterval=1) as file_progress, \
-            PotentiallySingleThreadedExecutor(max_workers=concurrency) as io_pool:
+            FutureFreeThreadPool(threads=concurrency) as io_pool:
         crawler = FileCrawler(input_paths, io_pool, file_progress)
         crawler.traverse()
 
@@ -271,13 +328,17 @@ def main(input_paths, output, verbose, progress, concurrency):
                 num_files,
                 sizeof_fmt(size))
     logger.info('Directory count: %d', crawler.num_directories)
+    logger.info('Number of candidate groups: %s (largest is %s files)',
+                len(size_groups), max((len(g) for _, g in size_groups), default=0))
     logger.info('Size filter reduced file count to: %d (%s)',
                 num_potential_dupes,
                 sizeof_fmt(size_potential_dupes))
     del crawler
 
-    with PotentiallySingleThreadedExecutor(max_workers=concurrency) as scheduler_pool, \
-            PotentiallySingleThreadedExecutor(max_workers=concurrency) as io_pool, \
+    size_groups.sort(key=lambda sg: sg[0] * len(sg[1]))
+
+    with FutureFreeThreadPool(threads=concurrency) as scheduler_pool, \
+            FutureFreeThreadPool(threads=concurrency) as io_pool, \
             tqdm(smoothing=0, desc='Filtering', unit=' files',
                  position=0, mininterval=1,
                  total=num_potential_dupes,
@@ -287,20 +348,28 @@ def main(input_paths, output, verbose, progress, concurrency):
                  total=size_potential_dupes, mininterval=1,
                  disable=not progress) as bytes_progress:
 
-        futures = []
-        dupe_finders = []
-        for s, group in size_groups:
-            dupe_finder = DupeFinder(size_bytes=s, pool=io_pool, output=output,
-                                     file_progress=file_progress, byte_progress=bytes_progress)
-            futures.append(scheduler_pool.submit(dupe_finder.find, group))
-            dupe_finders.append(dupe_finder)
+        size_num_dupes = []
+        dupe_finder = DupeFinder(pool=io_pool, output=output,
+                                 file_progress=file_progress, byte_progress=bytes_progress)
 
-        dupe_count = 0
-        dupe_total_size = 0
-        for dupe_finder, f in zip(dupe_finders, futures):
-            dupes = f.result()
-            dupe_count += len(dupes)
-            dupe_total_size += len(dupes) * dupe_finder.size_bytes
+        def callback(args):
+            size, dupes = args
+            size_num_dupes.append((size, len(dupes)))
+
+        def return_with_size(size_bytes, group):
+            dupes = dupe_finder.find(size_bytes, group)
+            return size_bytes, dupes
+
+        while size_groups:
+            size_bytes, group = size_groups.pop()
+            scheduler_pool.submit(return_with_size, size_bytes, group, callback=callback)
+        scheduler_pool.wait_until_complete()
+
+    dupe_count = 0
+    dupe_total_size = 0
+    for size_bytes, num_dupes in size_num_dupes:
+        dupe_count += num_dupes
+        dupe_total_size += num_dupes * size_bytes
 
     dt_complete = datetime.datetime.now()
     logger.info('Comparison time: %.1fs', (dt_complete - time_traverse).total_seconds())
